@@ -1,6 +1,6 @@
 /*
  * C++ BMP library.
- * Copyright (C) 2006 - 2009 René Rebe, ExactCODE GmbH Germany
+ * Copyright (C) 2006 - 2015 René Rebe, ExactCODE GmbH Germany
  * 
  * This program is free software; you can redistribute it and/or modify
  * it under the terms of the GNU General Public License as published by
@@ -32,6 +32,7 @@
 
 #include <string.h>
 #include <ctype.h>
+#include <vector>
 
 #include <Endianess.hh>
 #include <inttypes.h>
@@ -57,7 +58,7 @@ enum BMPType
     BMPT_WIN4,      /* BMP used in Windows 3.0/NT 3.51/95 */
     BMPT_WIN5,      /* BMP used in Windows NT 4.0/98/Me/2000/XP */
     BMPT_OS21,      /* BMP used in OS/2 PM 1.x */
-    BMPT_OS22       /* BMP used in OS/2 PM 2.x */
+    BMPT_OS22,      /* BMP used in OS/2 PM 2.x */
   };
 
 /*
@@ -83,16 +84,21 @@ enum BMPType
 
 enum BMPComprMethod
   {
-    BMPC_RGB = 0L,          /* Uncompressed */
-    BMPC_RLE8 = 1L,         /* RLE for 8 bpp images */
-    BMPC_RLE4 = 2L,         /* RLE for 4 bpp images */
-    BMPC_BITFIELDS = 3L,    /* Bitmap is not compressed and the colour table
+    BMPC_RGB = 0,          /* Uncompressed */
+    BMPC_RLE8 = 1,         /* RLE for 8 bpp images */
+    BMPC_RLE4 = 2,         /* RLE for 4 bpp images */
+    BMPC_BITFIELDS = 3,    /* Bitmap is not compressed and the colour table
 			     * consists of three DWORD color masks that specify
 			     * the red, green, and blue components of each
 			     * pixel. This is valid when used with
 			     * 16- and 32-bpp bitmaps. */
-    BMPC_JPEG = 4L,         /* Indicates that the image is a JPEG image. */
-    BMPC_PNG = 5L           /* Indicates that the image is a PNG image. */
+    BMPC_JPEG = 4,         /* Indicates that the image is a JPEG image. */
+    BMPC_PNG = 5,           /* Indicates that the image is a PNG image. */
+    BMPC_ALPHABITFIELDS = 6,
+    BMPC_CMYK = 11,
+    BMPC_CMYKRLE = 12,
+    BMPC_CMYRTLE = 13,
+    
   };
 
 enum BMPLCSType                 /* Type of logical color space. */
@@ -101,25 +107,30 @@ enum BMPLCSType                 /* Type of logical color space. */
 				 * gamma values are given in the appropriate
 				 * fields. */
     BMPLT_DEVICE_RGB = 1,
-    BMPLT_DEVICE_CMYK = 2
+    BMPLT_DEVICE_CMYK = 2,
   };
 
-typedef struct
+#ifdef _MSC_VER
+#pragma pack(push, 1)
+#endif
+
+struct BMPCIEXYZ
 {
   int32_t   iCIEX;
   int32_t   iCIEY;
   int32_t   iCIEZ;
-} BMPCIEXYZ;
+};
 
-typedef struct                  /* This structure contains the x, y, and z */
+struct BMPCIEXYZTriple          /* This structure contains the x, y, and z */
 {				/* coordinates of the three colors that */
 				/* correspond */
   BMPCIEXYZ   iCIERed;        /* to the red, green, and blue endpoints for */
   BMPCIEXYZ   iCIEGreen;      /* a specified logical color space. */
   BMPCIEXYZ   iCIEBlue;
-} BMPCIEXYZTriple;
+};
 
-typedef struct
+
+struct BMPFileHeader
 {
   char	bType[2];       /* Signature "BM" */
   EndianessConverter<uint32_t,LittleEndianTraits> iSize; /* Size in bytes of the bitmap file. Should
@@ -129,12 +140,14 @@ typedef struct
   uint16_t	iReserved1;     /* Reserved, set as 0 */
   uint16_t	iReserved2;     /* Reserved, set as 0 */
   EndianessConverter<uint32_t,LittleEndianTraits> iOffBits; /* Offset of the image from file start in bytes */
-}  __attribute__((packed)) BMPFileHeader;
+}
+#ifdef __GNUC__
+__attribute__((packed))
+#endif
+;
 
-/* File header size in bytes: */
-const int       BFH_SIZE = 14;
 
-typedef struct
+struct BMPInfoHeader
 {
   EndianessConverter<uint32_t,LittleEndianTraits> iSize; /* Size of BMPInfoHeader structure in bytes.
 				 * Should be used to determine start of the
@@ -183,36 +196,46 @@ typedef struct
 			 * in 16^16 format. */
   EndianessConverter<int32_t,LittleEndianTraits> iGammaGreen;    /* Toned response curve for green. */
   EndianessConverter<int32_t,LittleEndianTraits> iGammaBlue;     /* Toned response curve for blue. */
-}  __attribute__((packed)) BMPInfoHeader;
+}
+#ifdef __GNUC__
+__attribute__((packed))
+#endif
+;
 
 /*
  * Info header size in bytes:
  */
-static const unsigned int BIH_WIN4SIZE = 40; /* for BMPT_WIN4 */
-static const unsigned int BIH_WIN5SIZE = 57; /* for BMPT_WIN5 */
 static const unsigned int BIH_OS21SIZE = 12; /* for BMPT_OS21 */
 static const unsigned int BIH_OS22SIZE = 64; /* for BMPT_OS22 */
+static const unsigned int BIH_WIN4SIZE = 40; /* for BMPT_WIN4 */
+static const unsigned int BIH_WIN5SIZE = 56; /* for BMPT_WIN5 */
+static const unsigned int BIH_V4 = 108;
+static const unsigned int BIH_V5 = 124;
 
 /*
  * We will use plain byte array instead of this structure, but declaration
  * provided for reference
  */
-typedef struct
+struct BMPColorEntry
 {
   char       bBlue;
   char       bGreen;
   char       bRed;
   char       bReserved;      /* Must be 0 */
-} BMPColorEntry;
+};
+
+#ifdef _MSC_VER
+#pragma pack(pop)
+#endif
 
 /*
  * Image data in BMP file stored in BGR (or ABGR) format. We rearrange
  * pixels to RGB (RGBA) format.
  */
-static void
-rearrangePixels(uint8_t* buf, uint32_t width, uint32_t bit_count)
+static void rearrangePixels(uint8_t* buf, uint32_t width, uint32_t bit_count)
 {
-  char tmp;
+  uint16_t tmp;
+  uint16_t* buf16 = (uint16_t*)buf;
   
   switch (bit_count) {
     
@@ -226,7 +249,15 @@ rearrangePixels(uint8_t* buf, uint32_t width, uint32_t bit_count)
       *(buf + 2) = tmp;
     }
     break;
-  
+
+  case 48:
+    for (uint32_t i = 0; i < width; i++, buf16 += 3) {
+      tmp = *buf16;
+      *buf16 = *(buf16 + 2);
+      *(buf16 + 2) = tmp;
+    }
+    break;
+
   case 32:
     {
       uint8_t* buf1 = buf;
@@ -244,17 +275,9 @@ rearrangePixels(uint8_t* buf, uint32_t width, uint32_t bit_count)
   }
 }
 
-bool BMPCodec::readImage (std::istream* stream, Image& image, const std::string& decompres)
+int BMPCodec::readImage (std::istream* stream, Image& image, const std::string& decompres)
 {
   BMPFileHeader file_hdr;
-  BMPInfoHeader info_hdr;
-  enum BMPType bmp_type;
-
-  uint32_t row, stride;
-  
-  uint32_t clr_tbl_size = 0, n_clr_elems = 3;
-  uint8_t *clr_tbl = 0;
-  uint8_t* data = 0;
   
   stream->read ((char*)&file_hdr.bType, 2);
   if (file_hdr.bType[0] != 'B' || file_hdr.bType[1] != 'M') {
@@ -270,25 +293,55 @@ bool BMPCodec::readImage (std::istream* stream, Image& image, const std::string&
   
   // fix the iSize, in early BMP file this is pure garbage
   stream->seekg (0, std::ios::end);
-  file_hdr.iSize = stream->tellg ();
-  
+  file_hdr.iSize = stream->tellg (); // TODO: minus the header?
+
+  int i = readImageWithoutFileHeader(stream, image, decompres, &file_hdr);
+  return i;
+}
+
+int BMPCodec::readImageWithoutFileHeader (std::istream* stream, Image& image, const std::string& decompres, BMPFileHeader* _file_hdr)
+{
+  BMPFileHeader* file_hdr = _file_hdr;
+  BMPFileHeader file_header; // only used if no file_hdr is supplied
+  BMPInfoHeader info_hdr = {};
+  enum BMPType bmp_type;
+  int offset = file_hdr ? sizeof(*file_hdr) : 0;
+
+  uint32_t clr_tbl_size = 0, n_clr_elems = 3;
+  uint8_t* clr_tbl = 0;
+  uint8_t* data = 0;
+
   /* -------------------------------------------------------------------- */
   /*      Read the BMPInfoHeader.                                         */
   /* -------------------------------------------------------------------- */
   
-  stream->seekg (BFH_SIZE);
+  stream->seekg (offset);
   stream->read ((char*)&info_hdr.iSize, 4);
   
+   if (!_file_hdr) {
+    offset = 0;
+    file_hdr = &file_header;
+    stream->seekg (0, std::ios::end);
+    file_hdr->iSize = stream->tellg ();
+    file_hdr->iOffBits = info_hdr.iSize; // assumed to follow after info header
+
+    stream->seekg (offset + 4);
+  }
+  
   if (info_hdr.iSize == BIH_WIN4SIZE)
     bmp_type = BMPT_WIN4;
+  else if (info_hdr.iSize == BIH_WIN5SIZE)
+    bmp_type = BMPT_WIN5;
   else if (info_hdr.iSize == BIH_OS21SIZE)
     bmp_type = BMPT_OS21;
   else if (info_hdr.iSize == BIH_OS22SIZE || info_hdr.iSize == 16)
     bmp_type = BMPT_OS22;
-  else
+  else {
     bmp_type = BMPT_WIN5;
-  
-  if (bmp_type == BMPT_WIN4 || bmp_type == BMPT_WIN5 ||
+    std::cerr << "Unknown header size: " << info_hdr.iSize << std::endl;
+  }
+
+if (bmp_type == BMPT_WIN4 || bmp_type == BMPT_WIN5 ||
       bmp_type == BMPT_OS22) {
     stream->read((char*)&info_hdr.iWidth, 4);
     stream->read((char*)&info_hdr.iHeight, 4);
@@ -306,8 +359,8 @@ bool BMPCodec::readImage (std::istream* stream, Image& image, const std::string&
     stream->read((char*)&info_hdr.iAlphaMask, 4);
     
     n_clr_elems = 4;
-    image.setResolution((2.54 * info_hdr.iXPelsPerMeter + .05) / 100,
-                        (2.54 * info_hdr.iYPelsPerMeter + .05) / 100);
+    image.setResolution((2.54 * info_hdr.iXPelsPerMeter) / 100 + .5,
+                        (2.54 * info_hdr.iYPelsPerMeter) / 100 + .5);
   }
   
   if (bmp_type == BMPT_OS22) {
@@ -320,7 +373,6 @@ bool BMPCodec::readImage (std::istream* stream, Image& image, const std::string&
 
   if (bmp_type == BMPT_OS21) {
     int16_t  iShort;
-    
     stream->read ((char*)&iShort, 2);
     info_hdr.iWidth = iShort;
     stream->read ((char*)&iShort, 2);
@@ -341,15 +393,16 @@ bool BMPCodec::readImage (std::istream* stream, Image& image, const std::string&
   case 16:
   case 24:
   case 32:
+  case 48:
     break;
   default:
     std::cerr << "BMPCodec:: Cannot read " << info_hdr.iBitCount
 	      << " bit files." << std::endl;
     return false;
   }
-  
+
   image.w = info_hdr.iWidth;
-  image.h = (info_hdr.iHeight > 0) ? info_hdr.iHeight : -info_hdr.iHeight;
+  image.h = std::abs(info_hdr.iHeight); // negative when upside-down
   
   switch (info_hdr.iBitCount)
     {
@@ -366,17 +419,21 @@ bool BMPCodec::readImage (std::istream* stream, Image& image, const std::string&
 	  1 << image.bps : info_hdr.iClrUsed;
       else
 	clr_tbl_size = 1 << image.bps;
-      clr_tbl = (uint8_t *)
-	malloc (n_clr_elems * clr_tbl_size);
+
+      //std::cerr << "n_clr_elems: " << n_clr_elems << ", clr_tbl_size: " << clr_tbl_size << std::endl;
+
+      // if we had no file_hdr, update offset to compensate for color table
+      if (!_file_hdr) {
+        file_hdr->iOffBits = file_hdr->iOffBits + n_clr_elems * clr_tbl_size;
+      }
+
+      clr_tbl = (uint8_t *) malloc (n_clr_elems * clr_tbl_size);
       if (!clr_tbl) {
 	std::cerr << "BMPCodec:: Can't allocate space for color table" << std::endl;
-	goto bad;
+	return false;
       }
       
-      /*printf ("n_clr_elems: %d, clr_tbl_size: %d\n",
-	n_clr_elems, clr_tbl_size); */
-      
-      stream->seekg (BFH_SIZE + info_hdr.iSize);
+      stream->seekg (offset + info_hdr.iSize);
       stream->read ((char*)clr_tbl, n_clr_elems * clr_tbl_size);
       
       /*for(clr = 0; clr < clr_tbl_size; ++clr) {
@@ -398,12 +455,17 @@ bool BMPCodec::readImage (std::istream* stream, Image& image, const std::string&
       image.spp = 3;
       image.bps = 8;
       break;
-    
+
+    case 48:
+      image.spp = 3;
+      image.bps = 16;
+      break;
+
     default:
       break;
     }
   
-  stride = image.stride ();
+  uint32_t stride = image.stride ();
   /*printf ("w: %d, h: %d, spp: %d, bps: %d, colorspace: %d\n",
    *w, *h, *spp, *bps, info_hdr.iCompression); */
   
@@ -431,47 +493,45 @@ bool BMPCodec::readImage (std::istream* stream, Image& image, const std::string&
     {
       uint32_t file_stride = ((image.w * info_hdr.iBitCount + 7) / 8 + 3) / 4 * 4;
       
-      /*printf ("bitcount: %d, stride: %d, file stride: %d\n",
-	      info_hdr.iBitCount, stride, file_stride);
+      /*std::cerr << "bitcount: " << info_hdr.iBitCount << ", stride: " << stride
+                << ", file stride: " << file_stride << std::endl;
      
-      printf ("red mask: %x, green mask: %x, blue mask: %x\n",
-      info_hdr.iRedMask, info_hdr.iGreenMask, info_hdr.iBlueMask); */
-      
-      data = (uint8_t*) malloc (stride*image.h);
+      std::cerr << std::hex << "red mask: " << info_hdr.iRedMask
+                << ", green mask: " << info_hdr.iGreenMask
+                << ", blue mask: " << info_hdr.iBlueMask << std::dec << std::endl; */
+
+      image.resize(image.w, image.h);
+      data = image.getRawData();
       uint8_t* row_data = (uint8_t*) malloc (file_stride);
-      if (!data || !row_data) {
+      if (!row_data) {
 	std::cerr << "Can't allocate space for image buffer\n";
 	goto bad1;
       }
       
-      const int r_shift = last_bit_set (info_hdr.iRedMask) - 7;
-      const int g_shift = last_bit_set (info_hdr.iGreenMask) - 7;
-      const int b_shift = last_bit_set (info_hdr.iBlueMask) - 7;
-      
-      for (row = 0; row < (uint32_t)image.h; ++row) {
-	std::istream::pos_type offset;
-	
-	if (info_hdr.iHeight > 0)
-	  offset = file_hdr.iOffBits + (image.h - row - 1) * file_stride;
-	else
-	  offset = file_hdr.iOffBits + row * file_stride;
-	
+      for (uint32_t row = 0; row < (uint32_t)image.h; ++row)
+      {
+	std::istream::pos_type offset = file_hdr->iOffBits + row * file_stride;
 	stream->seekg (offset);
-	/*
-	  if (stream->tellg () != offset) {
-	  fprintf(stderr, "scanline %lu: Seek error\n", (unsigned long) row);
-	  }
-	*/
 	
-	if (stream->read ((char*)row_data, file_stride) < 0) {
-	  std::cerr << "scanline " << row << ": Read error\n";
+	if (stream->tellg () != offset) {
+	  std::cerr << "scanline " << row << " Seek error: " << stream->tellg() << " vs " << offset << std::endl;
 	}
 	
-	// convert to RGB
-	if (info_hdr.iCompression == BMPC_BITFIELDS)
+	stream->read((char*)row_data, file_stride);
+	if (!stream->good()) {
+	  std::cerr << "bmp read error: scanline " << row << "\n";
+	} else {
+	  // convert to RGB
+	  uint8_t* rgb_ptr = data + stride * (info_hdr.iHeight < 0 ? row : image.h - row - 1);
+
+	  if (info_hdr.iCompression == BMPC_BITFIELDS)
 	  {
 	    uint8_t* bf_ptr = row_data;
-	    uint8_t* rgb_ptr = data + stride * row;
+
+	    const int r_shift = last_bit_set (info_hdr.iRedMask) - 7;
+	    const int g_shift = last_bit_set (info_hdr.iGreenMask) - 7;
+	    const int b_shift = last_bit_set (info_hdr.iBlueMask) - 7;
+	    const int a_shift = last_bit_set (info_hdr.iAlphaMask) - 7;
 	    
 	    for (int i = 0; i < image.w; ++i, rgb_ptr += 3)
 	      {
@@ -491,11 +551,15 @@ bool BMPCodec::readImage (std::istream* stream, Image& image, const std::string&
 		  rgb_ptr[2] = (val & info_hdr.iBlueMask) >> b_shift;
 		else
 		  rgb_ptr[2] = (val & info_hdr.iBlueMask) << -b_shift;
+		/*if (a_shift > 0)
+		  rgb_ptr[3] = (val & info_hdr.iAlphaMask) >> a_shift;
+		else
+		rgb_ptr[3] = (val & info_hdr.iAlphaMask) << -a_shift;*/
 	      }
+	  } else {
+	    rearrangePixels (row_data, image.w, info_hdr.iBitCount);
+	    memcpy (rgb_ptr, row_data, stride);
 	  }
-	else {
-	  rearrangePixels (row_data, image.w, info_hdr.iBitCount);
-	  memcpy (data+stride*row, row_data, stride);
 	}
       }
       free(row_data);
@@ -516,7 +580,7 @@ bool BMPCodec::readImage (std::istream* stream, Image& image, const std::string&
       //std::cerr << "RLE" << (info_hdr.iCompression == BMPC_RLE4 ? "4" : "8")
       //	<< " compressed\n";
       
-      compr_size = file_hdr.iSize - file_hdr.iOffBits;
+      compr_size = file_hdr->iSize - file_hdr->iOffBits;
       uncompr_size = image.w * image.h;
       
       comprbuf = (uint8_t *) malloc( compr_size );
@@ -530,7 +594,7 @@ bool BMPCodec::readImage (std::istream* stream, Image& image, const std::string&
 	goto bad1;
       }
       
-      stream->seekg (*file_hdr.iOffBits);
+      stream->seekg (*file_hdr->iOffBits);
       stream->read ((char*)comprbuf, compr_size);
       i = j = x = 0;
       
@@ -555,7 +619,7 @@ bool BMPCodec::readImage (std::istream* stream, Image& image, const std::string&
 	  i++;
 	  if ( comprbuf[i] == 0 ) {         /* Next scanline */
 	    i++;
-	    x = 0;;
+	    x = 0;
 	  }
 	  else if ( comprbuf[i] == 1 )    /* End of image */
 	    break;
@@ -597,7 +661,7 @@ bool BMPCodec::readImage (std::istream* stream, Image& image, const std::string&
       }
       
       // TODO: suboptimal, later improve the above to yield the corrent orientation natively
-      for (row = 0; row < (uint32_t)image.h; ++row) {
+      for (uint32_t row = 0; row < (uint32_t)image.h; ++row) {
 	memcpy (data + row * image.w, uncomprbuf + (image.h - 1 - row) * image.w, image.w);
 	rearrangePixels(data + row * image.w, image.w, info_hdr.iBitCount);
       }
@@ -605,21 +669,20 @@ bool BMPCodec::readImage (std::istream* stream, Image& image, const std::string&
       free(uncomprbuf);
       image.bps = 8;
     }
+    image.setRawData(data);
     break;
   } /* switch */
   
-  image.setRawData (data);
-  
-  // convert to RGB color-space - we do not handle palete images internally
+  // convert to RGB color-space - we do not handle palette images internally
   
   // no color table anyway or RGB* ?
   if (clr_tbl && image.spp < 3)
     {
-      uint16_t* rmap = new uint16_t [1 << image.bps];
-      uint16_t* gmap = new uint16_t [1 << image.bps];
-      uint16_t* bmap = new uint16_t [1 << image.bps];
+      uint16_t* rmap = new uint16_t [clr_tbl_size];
+      uint16_t* gmap = new uint16_t [clr_tbl_size];
+      uint16_t* bmap = new uint16_t [clr_tbl_size];
       
-      for (int i = 0; i < (1 << image.bps); ++i) {
+      for (unsigned int i = 0; i < clr_tbl_size; ++i) {
 	// BMP maps have BGR order ...
 	rmap[i] = 0x101 * clr_tbl[i * n_clr_elems + 2];
 	gmap[i] = 0x101 * clr_tbl[i * n_clr_elems + 1];
@@ -628,9 +691,9 @@ bool BMPCodec::readImage (std::istream* stream, Image& image, const std::string&
       
       colorspace_de_palette (image, clr_tbl_size, rmap, gmap, bmap);
       
-      delete (rmap);
-      delete (gmap);
-      delete (bmap);
+      delete[] (rmap);
+      delete[] (gmap);
+      delete[] (bmap);
       
       free(clr_tbl);
       clr_tbl = NULL;
@@ -644,28 +707,24 @@ bool BMPCodec::readImage (std::istream* stream, Image& image, const std::string&
   clr_tbl = NULL;
   
   return false;
-  
- bad:
-  return false;
 }
 
 bool BMPCodec::writeImage (std::ostream* stream, Image& image, int quality,
 			   const std::string& compress)
 {
-  if (image.bps > 8 || image.bps == 2 || image.spp > 3) {
+  const int hdr_size = image.spp == 4 ? BIH_WIN5SIZE : BIH_WIN4SIZE;
+  const unsigned stride = image.stride();
+  const int n_clr_elems = 4; // we write "modern" formats, not the vintage OS/2 flavour
+  
+  if (image.bps > 16 || image.bps == 2 || image.spp > 4) {
     std::cerr << "BMPCodec: " << image.bps << " bits and "
 	      << image.spp << " samples not supported." << std::endl;
     return false;
   }
   
-  const int n_clr_elems = 4; // we write "modern" formats, not the first OS2 flavours
-  
   BMPFileHeader file_hdr;
   BMPInfoHeader info_hdr;
   
-  int hdr_size = BIH_WIN4SIZE;
-  int stride = image.stride ();
-  
   memset (&file_hdr, 0, sizeof (file_hdr));
   memset (&info_hdr, 0, sizeof (info_hdr));
 
@@ -683,8 +742,8 @@ bool BMPCodec::writeImage (std::ostream* stream, Image& image, int quality,
   info_hdr.iBitCount = image.spp * image.bps;
   info_hdr.iCompression = BMPC_RGB;
   info_hdr.iSizeImage  = image.stride()*image.h; // TODO: compressed size
-  info_hdr.iXPelsPerMeter = (int32_t) (image.resolutionX() * 100 / 2.54);
-  info_hdr.iYPelsPerMeter = (int32_t) (image.resolutionY() * 100 / 2.54);
+  info_hdr.iXPelsPerMeter = (int32_t) (100. * image.resolutionX() / 2.54 + .5);
+  info_hdr.iYPelsPerMeter = (int32_t) (100. * image.resolutionY() / 2.54 + .5);
   info_hdr.iClrUsed = image.spp == 1 ? 1 << image.bps : 0;
   info_hdr.iClrImportant = 0;
   info_hdr.iRedMask = 0;
@@ -695,19 +754,22 @@ bool BMPCodec::writeImage (std::ostream* stream, Image& image, int quality,
   // BMP image payload needs to be 4 byte aligned :-(
   int file_stride = ((image.w * info_hdr.iBitCount + 7) / 8 + 3) / 4 * 4;
   
-  file_hdr.iOffBits = BFH_SIZE + hdr_size + info_hdr.iClrUsed * n_clr_elems;
+  file_hdr.iOffBits = sizeof(file_hdr) + hdr_size + info_hdr.iClrUsed * n_clr_elems;
   file_hdr.iSize =  file_hdr.iOffBits + file_stride * image.h;
   
   // write header meta info
-  stream->write ((char*)&file_hdr, BFH_SIZE);
+  stream->write ((char*)&file_hdr, sizeof(file_hdr));
   stream->write ((char*)&info_hdr, hdr_size);
   
   // write color table
   if (info_hdr.iClrUsed) {
     int n = info_hdr.iClrUsed;
-    
+#ifdef _MSC_VER
+    std::vector<uint8_t> _clrtbl(n_clr_elems*n);
+    uint8_t* clrtbl = &_clrtbl[0];
+#else
     uint8_t clrtbl [n_clr_elems*n];
-    
+#endif
     for (int i = 0; i < n; ++i) {
       clrtbl[n_clr_elems*i+0] = clrtbl[n_clr_elems*i+1] = clrtbl[n_clr_elems*i+2] = i * 0xff / (n - 1);
       
@@ -721,15 +783,20 @@ bool BMPCodec::writeImage (std::ostream* stream, Image& image, int quality,
   switch (info_hdr.iCompression) {
   case BMPC_RGB:
     {
+#ifdef _MSC_VER
+      std::vector<uint8_t> _payload(file_stride);
+      uint8_t* payload = &_payload[0];
+#else
       uint8_t payload [file_stride];
-      
+#endif
       for (int row = image.h-1; row >=0; --row)
 	{
 	  memcpy (payload, image.getRawData() + stride*row, stride);
 	  rearrangePixels (payload, image.w, info_hdr.iBitCount);
 	  
-	  if (!stream->write ((char*)payload, file_stride) ) {
+	  if (!stream->write ((char*)payload, file_stride)) {
 	    std::cerr << "scanline " << row << " write error" << std::endl;
+	    return false;
 	  }
 	}
     }
