@@ -1,6 +1,6 @@
 /*
- * Colorspace conversions..
- * Copyright (C) 2006 - 2009 René Rebe, ExactCOD GmbH Germany
+ * Colorspace conversions.
+ * Copyright (C) 2006 - 2015 René Rebe, ExactCOD GmbH Germany
  * Copyright (C) 2007 Susanne Klaus, ExactCODE
  *
  * This program is free software; you can redistribute it and/or modify
@@ -19,6 +19,7 @@
 
 #include <iostream>
 #include <map>
+#include <vector>
 
 #include "Image.hh"
 #include "ImageIterator2.hh"
@@ -28,13 +29,53 @@
 
 #include "Endianess.hh"
 
+template <typename T>
+struct histogram_template
+{
+  std::vector<std::vector<unsigned int> > operator() (Image& image, int bins = 256)
+  {
+    std::vector<std::vector<unsigned int> > hist;
+    hist.resize(image.spp);
+
+    typename T::accu a;
+    
+      for (int i = 0; i < image.spp; ++i)
+	  hist[i].resize(bins, 0);
+      
+      typename T::accu one = T::accu::one();
+
+      T it (image);
+      for (int y = 0; y < image.h; ++y) {
+	for (int x = 0; x < image.w; ++x, ++it)
+	  {
+	    a = *it;
+            for (int i = 0; i < image.spp; ++i) {
+		typename T::accu::vtype v = a.v[i];
+    		int j = (int)(v * (bins-1) / one.v[i]);
+		if (j < 0) j = 0;
+		else if (j >= bins) j = bins - 1;
+		hist[i][j]++;
+	    }
+	  }
+      }
+
+      return hist;
+    }
+};
+
+std::vector<std::vector<unsigned int> > histogram(Image& image, int bins)
+{
+    return codegen_return<std::vector<std::vector<unsigned int> >,
+	histogram_template> (image, bins);
+}
+
 template <typename T>
 struct normalize_template
 {
   void operator() (Image& image, uint8_t l, uint8_t h)
   {
     typename T::accu a;
-    typename T::accu::vtype black, white;
+    typename T::accu::vtype black(0), white(0);
     
     // darkest 1%, lightest .5%
     const int white_point = image.w * image.h / 100;
@@ -45,7 +86,7 @@ struct normalize_template
       typedef std::vector<typename T::accu::vtype> histogram_type;
       typename T::accu::vtype hsize;
       T::accu::one().getL(hsize);
-      histogram_type histogram(hsize);
+      histogram_type histogram(hsize + 1);
       
       T it (image);
       typename T::accu::vtype l;
@@ -115,39 +156,102 @@ void normalize (Image& image, uint8_t l, uint8_t h)
 
 void colorspace_rgba8_to_rgb8 (Image& image)
 {
-  uint8_t* output = image.getRawData();
-  for (uint8_t* it = image.getRawData(); it < image.getRawData() + image.w*image.h*image.spp;)
+  unsigned ostride = image.stride();
+  image.spp = 3; image.rowstride = 0;
+  
+  for (int y = 0; y < image.h; ++y)
+  {
+    uint8_t* output = image.getRawData() + y * image.stride();
+    uint8_t* it = image.getRawData() + y * ostride;
+    for (int x = 0; x < image.w; ++x)
     {
       *output++ = *it++;
       *output++ = *it++;
       *output++ = *it++;
       it++; // skip over a
     }
-  image.spp = 3; // converted data right now
-  image.setRawData();
+  }
+
+  image.resize(image.w, image.h); // realloc
 }
 
-void colorspace_rgb8_to_gray8 (Image& image)
+void colorspace_argb8_to_rgb8 (Image& image)
 {
-  uint8_t* output = image.getRawData();
-  for (uint8_t* it = image.getRawData(); it < image.getRawData() + image.stride() * image.h;)
+  uint8_t* data = image.getRawData();
+  unsigned ostride = image.stride();
+  
+  image.spp = 3; image.rowstride = 0;
+  for (int y = 0; y < image.h; ++y)
+  {
+    uint8_t* output = data + y * image.stride();
+    uint8_t* it = data + y * ostride;
+    for (int x = 0; x < image.w; ++x)
+    {
+      it++; // skip over a
+      *output++ = *it++;
+      *output++ = *it++;
+      *output++ = *it++;
+    }
+  }
+
+  image.resize(image.w, image.h); // realloc
+}
+
+void colorspace_cmyk_to_rgb8 (Image& image)
+{
+  uint8_t* data = image.getRawData();
+  unsigned ostride = image.stride();
+  
+  image.spp = 3; image.rowstride = 0;
+  for (int y = 0; y < image.h; ++y)
+  {
+    uint8_t* output = data + y * image.stride();
+    uint8_t* it = data + y * ostride;
+    for (int x = 0; x < image.w; ++x, it += 4, output += 3)
+    {
+      uint8_t c = it[0], m = it[1], y = it[2], k = it[3]; 
+      output[0] = 0xff - std::min(c+k, 0xff); // ((0xff-c)*(0xff-k)) >> 8;
+      output[1] = 0xff - std::min(m+k, 0xff); // ((0xff-m)*(0xff-k)) >> 8;
+      output[2] = 0xff - std::min(y+k, 0xff); // ((0xff-y)*(0xff-k)) >> 8;
+    }
+  }
+
+  image.resize(image.w, image.h); // realloc
+}
+
+void colorspace_rgb8_to_gray8 (Image& image, const int bytes, const int wR, const int wG, const int wB)
+{
+  unsigned ostride = image.stride();
+  image.spp = 1; image.rowstride = 0;
+  
+  const int sum = wR + wG + wB;
+  for (int y = 0; y < image.h; ++y)
+  {
+    uint8_t* output = image.getRawData() + y * image.stride();
+    uint8_t* it = image.getRawData() + y * ostride;
+    for (int x = 0; x < image.w; ++x, it += bytes)
     {
       // R G B order and associated weighting
-      int c = (int)*it++ * 28;
-      c += (int)*it++ * 59;
-      c += (int)*it++ * 11;
-      
-      *output++ = (uint8_t)(c / 100);
+      int c  = (int)it[0] * wR;
+      c += (int)it[1] * wG;
+      c += (int)it[2] * wB;
+      *output++ = (uint8_t)(c / sum);
     }
-  image.spp = 1; // converted data right now
-  image.setRawData();
+  }
+  
+  image.resize(image.w, image.h); // realloc
 }
 
 void colorspace_rgb16_to_gray16 (Image& image)
 {
-  uint16_t* output = (uint16_t*)image.getRawData();
-  for (uint16_t* it = output;
-       it < (uint16_t*)(image.getRawData() + image.stride() * image.h);)
+  unsigned ostride = image.stride();
+  image.spp = 1; image.rowstride = 0;
+
+  for (int y = 0; y < image.h; ++y)
+  {
+    uint16_t* output = (uint16_t*)(image.getRawData() + y * image.stride());
+    uint16_t* it = (uint16_t*)image.getRawData() + y * ostride;
+    for (int x = 0; x < image.w; ++x)
     {
       // R G B order and associated weighting
       int c = (int)*it++ * 28;
@@ -156,13 +260,14 @@ void colorspace_rgb16_to_gray16 (Image& image)
       
       *output++ = (uint16_t)(c / 100);
     }
-  image.spp = 1; // converted data right now
-  image.setRawData();
+  }
+  
+  image.resize(image.w, image.h, image.stride()); // realloc
 }
 
 void colorspace_rgb8_to_rgb8a (Image& image, uint8_t alpha)
 {
-  image.setRawDataWithoutDelete	((uint8_t*)realloc(image.getRawData(),
+  image.setRawDataWithoutDelete ((uint8_t*)realloc(image.getRawData(),
 						   image.w * 4 * image.h));
   image.setSamplesPerPixel(4);
   
@@ -179,26 +284,30 @@ void colorspace_rgb8_to_rgb8a (Image& image, uint8_t alpha)
 
 void colorspace_gray8_threshold (Image& image, uint8_t threshold)
 {
-  for (uint8_t* it = image.getRawData(); it < image.getRawDataEnd(); ++it)
-    *it = *it > threshold ? 0xFF : 0x00;
-  
+   for (int y = 0; y < image.h; ++y)
+  {
+    uint8_t* it = image.getRawData() + y * image.stride();
+    for (int x = 0; x < image.w; ++x)
+     *it = *it > threshold ? 0xFF : 0x00;
+  }
+
   image.setRawData();
 }
 
-void colorspace_gray8_denoise_neighbours (Image &image)
+void colorspace_gray8_denoise_neighbours (Image &image, bool gross)
 {
-  // we need some pixels to compare, also avoids conditionals
-  // below
-  if (image.w < 3 ||
-      image.h < 3)
+  if (image.bps != 8 || image.spp != 1)
     return;
   
-  uint8_t* it = image.getRawData();
+  unsigned stride = image.stride();
+  uint8_t* data = image.getRawData();
+  uint8_t* ndata = (uint8_t*)malloc(stride * image.h);
   
   struct compare_and_set
   {
     const Image& image;
-    const unsigned int stride;
+    const int stride;
+    
     compare_and_set (const Image& _image)
       : image(_image), stride (image.stride())
     {
@@ -206,56 +315,92 @@ void colorspace_gray8_denoise_neighbours (Image &image)
     
     // without the inner(area) compiler guidance the conditionals are
     // not optimized away well enough
-    void operator() (const int x, const int y, uint8_t* it,
-		     const bool inner = false)
+    void operator() (const int x, const int y, uint8_t* it, uint8_t* it2,
+		     const bool inner, const bool gross = false)
     {
-      unsigned int n = 0, sum = 0;
+      int n = 0, sum = 0;
       
+      // +
       if (inner || x > 0)
-	sum += *(it-1), ++n;
+	sum += it[-1], ++n;
       if (inner || y > 0)
-	sum += *(it-stride), ++n;
+	sum += it[-stride], ++n;
+      
       if (inner || x < image.w-1)
-	sum += *(it+1), ++n;
+	sum += it[1], ++n;
       if (inner || y < image.h-1)
-	sum += *(it+stride), ++n;
+	sum += it[stride], ++n;
+
+      // x
+      if (gross) {
+	if (inner || y > 0) {
+	  if (inner || x > 0)
+	    sum += it[-stride - 1], ++n;
+	  if (inner || x < image.w-1)
+	    sum += it[-stride + 1], ++n;
+	}
+	
+	if (inner || y < image.h-1) {
+	  if (inner || x > 0)
+	    sum += it[stride - 1], ++n;
+	  if (inner || x < image.w-1)
+	    sum += it[stride + 1], ++n;
+	}
+      }
       
       // if all direct neighbours are black or white, fill it
-      if (sum == 0)
-	*it = 0;
-      else if (sum == n * 0xff)
-	*it = 0xff;
+      if (gross) {
+        if (sum <= 1 * 0xff)
+	  *it2 = 0;
+        else if (sum >= (n - 1) * 0xff)
+	  *it2 = 0xff;
+	else
+	  *it2 = *it;
+      } else {
+        if (sum == 0)
+	  *it2 = 0;
+        else if (sum == n * 0xff)
+	  *it2 = 0xff;
+	else
+	  *it2 = *it;
+      }
     }
   } compare_and_set (image);
   
   for (int y = 0; y < image.h; ++y)
     {
+      uint8_t* it = data + y * stride;
+      uint8_t* it2 = ndata + y * stride;
+	  
       // optimize conditionals away for the inner area
       if (y > 0 && y < image.h-1)
 	{
-	  compare_and_set (1, y, it++);
-	  for (int x = 1; x < image.w-1; ++x, ++it)
-	    compare_and_set (x, y, it, true);
-	  compare_and_set (image.w-1, y, it++);
+	  compare_and_set (0, y, it++, it2++, false, gross);
+	  for (int x = 1; x < image.w-1; ++x)
+	    compare_and_set (x, y, it++, it2++, true, gross);
+	  compare_and_set (image.w-1, y, it++, it2++, false, gross);
 	}
       else // quite some out of bounds conditions to check
-	for (int x = 0; x < image.w; ++x, ++it) 
-	  compare_and_set (x, y, it);
+	for (int x = 0; x < image.w; ++x) 
+	  compare_and_set (x, y, it++, it2++, false, gross);
     }
     
-  image.setRawData();
+  image.setRawData(ndata);
 }
 
 void colorspace_gray8_to_gray1 (Image& image, uint8_t threshold)
 {
-  uint8_t *output = image.getRawData();
-  uint8_t *input = image.getRawData();
-  
+  unsigned ostride = image.stride();
+  image.bps = 1; image.rowstride = 0;
+
   for (int row = 0; row < image.h; row++)
     {
+      uint8_t *output = image.getRawData() + row * image.stride();
+      uint8_t *input = image.getRawData() + row * ostride;
+
       uint8_t z = 0;
       int x = 0;
-      for (; x < image.w; x++)
+      for (; x < image.w; ++x)
 	{
 	  z <<= 1;
 	  if (*input++ > threshold)
@@ -274,17 +419,20 @@ void colorspace_gray8_to_gray1 (Image& image, uint8_t threshold)
 	  *output++ = z;
 	}
     }
-  image.bps = 1;
-  image.setRawData();
+  
+  image.resize(image.w, image.h); // realloc
 }
 
 void colorspace_gray8_to_gray4 (Image& image)
 {
-  uint8_t *output = image.getRawData();
-  uint8_t *input = image.getRawData();
+  unsigned ostride = image.stride();
+  image.bps = 4; image.rowstride = 0;
   
   for (int row = 0; row < image.h; row++)
     {
+      uint8_t *output = image.getRawData() + row * image.stride();
+      uint8_t *input = image.getRawData() + row * ostride;
+  
       uint8_t z = 0;
       int x = 0;
       for (; x < image.w; x++)
@@ -305,16 +453,20 @@ void colorspace_gray8_to_gray4 (Image& image)
 	  *output++ = z;
 	}
     }
-  image.bps = 4;
-  image.setRawData();
+  
+  image.resize(image.w, image.h); // realloc
 }
+
 void colorspace_gray8_to_gray2 (Image& image)
 {
-  uint8_t *output = image.getRawData();
-  uint8_t *input = image.getRawData();
+  unsigned ostride = image.stride();
+  image.bps = 2; image.rowstride = 0;
   
   for (int row = 0; row < image.h; ++row)
     {
+      uint8_t *output = image.getRawData() + row * image.stride();
+      uint8_t *input = image.getRawData() + row * ostride;
+
       uint8_t z = 0;
       int x = 0;
       for (; x < image.w; x++)
@@ -335,32 +487,40 @@ void colorspace_gray8_to_gray2 (Image& image)
 	  *output++ = z;
 	}
     }
-  image.bps = 2;
-  image.setRawData();
+  
+  image.resize(image.w, image.h); // realloc
 }
 
 void colorspace_gray8_to_rgb8 (Image& image)
 {
-  uint8_t* data = (uint8_t*)malloc (image.w*image.h*3);
-  uint8_t* output = data;
-  for (uint8_t* it = image.getRawData ();
-       it < image.getRawData() + image.w*image.h*image.spp; ++it)
+  const int stride = image.stride();
+  const int nstride = image.w * 3;
+  image.setRawDataWithoutDelete((uint8_t*)realloc(image.getRawData(),
+						  std::max(stride, nstride) * image.h));
+  uint8_t* data = image.getRawData();
+  uint8_t* output = data + image.h * nstride - 1;
+  for (int y = image.h - 1; y >= 0; --y)
     {
-      *output++ = *it;
-      *output++ = *it;
-      *output++ = *it;
+      uint8_t* it = data + y * stride;
+      for (int x = image.w - 1; x >= 0; --x)
+	{
+	  *output-- = it[x];
+	  *output-- = it[x];
+	  *output-- = it[x];
+	}
     }
-  image.spp = 3; // converted data right now
-  image.setRawData(data);
+  
+  image.spp = 3;
+  image.resize(image.w, image.h); // realloc
 }
 
 void colorspace_grayX_to_gray8 (Image& image)
 {
   uint8_t* old_data = image.getRawData();
-  int old_stride = image.stride();
+  unsigned old_stride = image.stride();
   
   const int bps = image.bps;
-  image.bps = 8;
+  image.bps = 8; image.rowstride = 0;
   image.setRawDataWithoutDelete ((uint8_t*)malloc(image.h * image.stride()));
   uint8_t* output = image.getRawData();
   
@@ -378,9 +538,8 @@ void colorspace_grayX_to_gray8 (Image& image)
   const unsigned int bitshift = 8 - bps;
   for (int row = 0; row < image.h; ++row)
     {
-      uint8_t* input  = old_data + row * old_stride;
-      uint8_t z = 0;
-      unsigned int bits = 0;
+      uint8_t* input = old_data + row * old_stride;
+      uint8_t z = 0, bits = 0;
       
       for (int x = 0; x < image.w; ++x)
 	{
@@ -423,7 +582,7 @@ void colorspace_grayX_to_rgb8 (Image& image)
   const unsigned int bitshift = 8 - bps;
   for (int row = 0; row < image.h; ++row)
     {
-      uint8_t* input  = old_data + row * old_stride;
+      uint8_t* input = old_data + row * old_stride;
       uint8_t z = 0;
       unsigned int bits = 0;
       
@@ -451,14 +610,15 @@ void colorspace_gray1_to_gray2 (Image& image)
   int old_stride = image.stride();
   
   image.bps = 2;
-  image.setRawDataWithoutDelete ((uint8_t*) malloc (image.h*image.stride()));
+  image.rowstride = 0;
+  image.setRawDataWithoutDelete ((uint8_t*)malloc(image.h * image.stride()));
   uint8_t* output = image.getRawData();
   
   for (int row = 0; row < image.h; ++row)
     {
       uint8_t z = 0;
       uint8_t zz = 0;
-      uint8_t* input  = old_data + row * old_stride;
+      uint8_t* input = old_data + row * old_stride;
 
       int x;
       for (x = 0; x < image.w; ++x)
@@ -491,7 +651,7 @@ void colorspace_gray1_to_gray4 (Image& image)
   int old_stride = image.stride();
   
   image.bps = 4;
-  image.setRawDataWithoutDelete ((uint8_t*) malloc (image.h*image.stride()));
+  image.setRawDataWithoutDelete ((uint8_t*)malloc(image.h * image.stride()));
   uint8_t* output = image.getRawData();
   
   for (int row = 0; row < image.h; ++row)
@@ -499,7 +659,7 @@ void colorspace_gray1_to_gray4 (Image& image)
       uint8_t z = 0;
       uint8_t zz = 0;
       
-      uint8_t* input  = old_data + row * old_stride;
+      uint8_t* input = old_data + row * old_stride;
       
       int x;
       for (x = 0; x < image.w; ++x)
@@ -530,41 +690,48 @@ void colorspace_gray1_to_gray4 (Image& image)
 void colorspace_16_to_8 (Image& image)
 {
   uint8_t* output = image.getRawData();
-  for (uint8_t* it = output; it < image.getRawDataEnd(); it += 2)
+  unsigned ostride = image.stride();
+  image.bps = 8; image.rowstride = 0;
+  
+  for (int y = 0; y < image.h; ++y)
     {
-      if (Exact::NativeEndianTraits::IsBigendian)
-	*output++ = it[0];
-      else
-	*output++ = it[1];
+      uint16_t* it = (uint16_t*)(image.getRawData() + y * ostride);
+      for (int x = 0; x < image.stride(); ++x)
+	{
+	  *output++ = it[x] >> 8;
+	}
     }
-  image.bps = 8; // converted 8bit data
-
-  // reallocate, to free half of the memory
-  image.setRawDataWithoutDelete	((uint8_t*)realloc(image.getRawData(),
-		    		  image.stride() * image.h));
+  
+  image.resize(image.w, image.h); // realloc
 }
 
 void colorspace_8_to_16 (Image& image)
 {
-  image.setRawDataWithoutDelete	((uint8_t*)realloc(image.getRawData(),
-		    		 image.stride() * 2 * image.h));
-	
-  uint8_t* data = image.getRawData();
-  uint16_t* data16 = (uint16_t*) data;
+  image.setRawDataWithoutDelete((uint8_t*)realloc(image.getRawData(),
+						  image.stride() * 2 * image.h));
   
-  for (signed int i = image.stride() * image.h - 1; i >= 0; --i)
+  uint8_t* data = image.getRawData();
+  unsigned stride = image.stride();
+  for (int y = image.h - 1; y >= 0; --y)
     {
-      data16[i] = data[i] * 0xffff / 255;
+      uint8_t* data8 = data + y * stride;
+      uint16_t* data16 = (uint16_t*)(data + y * stride * 2);
+      
+      for (int x = image.stride() - 1; x >= 0; --x)
+	{
+	  data16[x] = data8[x] * 0xffff / 255;
+	}
     }
   
+  image.rowstride *= 2;
   image.bps = 16; // converted 16bit data
 }
 
 void colorspace_de_palette (Image& image, int table_entries,
-			    uint16_t* rmap, uint16_t* gmap, uint16_t* bmap)
+			    uint16_t* rmap, uint16_t* gmap, uint16_t* bmap, uint16_t* amap)
 {
   // detect 1bps b/w tables
-  if (image.bps == 1) {
+  if (image.bps == 1 && table_entries >= 2 && !amap) {
     if (rmap[0] == 0 &&
 	gmap[0] == 0 &&
 	bmap[0] == 0 &&
@@ -594,7 +761,7 @@ void colorspace_de_palette (Image& image, int table_entries,
   
   // detect gray tables
   bool is_gray = false;
-  if (table_entries > 1) {
+  if (table_entries > 1 && !amap) {
     bool is_ordered_gray = (image.bps == 8 || image.bps == 4 ||
 			    image.bps == 2) && (1 << image.bps == table_entries);
     is_gray = true;
@@ -624,7 +791,9 @@ void colorspace_de_palette (Image& image, int table_entries,
   }
   
   int new_size = image.w * image.h;
-  if (!is_gray) // RGB
+  if (amap)
+    new_size *= 4; // RGBA, CMYK
+  else if (!is_gray) // RGB
     new_size *= 3;
   
   uint8_t* orig_data = image.getRawData();
@@ -635,35 +804,38 @@ void colorspace_de_palette (Image& image, int table_entries,
 
   // TODO: allow 16bit output if the palette contains that much dynamic
 
-  int bits_used = 0;
-  int x = 0;
+  const unsigned int bitshift = 8 - image.bps;
   while (dst < new_data + new_size)
-    {
-      uint8_t v = *src >> (8 - image.bps);
-      if (is_gray) {
-	*dst++ = rmap[v] >> 8;
-      } else {
-	*dst++ = rmap[v] >> 8;
-	*dst++ = gmap[v] >> 8;
-	*dst++ = bmap[v] >> 8;
-      }
-      
-      bits_used += image.bps;
-      ++x;
+  {
+    uint8_t z = 0;
+    unsigned int bits = 0;
 
-      if (bits_used == 8 || x == image.w) {
-	++src;
-	bits_used = 0;
-	if (x == image.w)
-	  x = 0;
+    for (int x = 0; x < image.w; ++x)
+    {
+      if (bits == 0) {
+        z = *src++;
+        bits = 8;
       }
-      else {
-	*src <<= image.bps;
+
+      if (is_gray) {
+	*dst++ = rmap[z >> bitshift] >> 8;
+      } else {
+	*dst++ = rmap[z >> bitshift] >> 8;
+	*dst++ = gmap[z >> bitshift] >> 8;
+	*dst++ = bmap[z >> bitshift] >> 8;
+	if (amap) *dst++ = amap[z >> bitshift] >> 8;
       }
+
+      z <<= image.bps;
+      bits -= image.bps;
     }
+  }
+
   image.bps = 8;
   if (is_gray)
     image.spp = 1;
+  else if (amap)
+    image.spp = 4;
   else
     image.spp = 3;
 
@@ -675,13 +847,7 @@ bool colorspace_by_name (Image& image, const std::string& target_colorspace,
 {
   std::string space = target_colorspace;
   std::transform (space.begin(), space.end(), space.begin(), tolower);
-  
-  // thru the codec?
-  if (!image.isModified() && image.getCodec())
-    if (space == "gray" || space == "gray8")
-      if (image.getCodec()->toGray(image))
-	return true;
-  
+    
   int spp, bps;
   if (space == "bw" || space == "bilevel" || space == "gray1") {
     spp = 1; bps = 1;
@@ -695,7 +861,7 @@ bool colorspace_by_name (Image& image, const std::string& target_colorspace,
     spp = 1; bps = 16;
   } else if (space == "rgb" || space == "rgb8") {
     spp = 3; bps = 8;
-  } else if (space == "rgba" || space == "rgb8a") {
+  } else if (space == "rgba" || space == "rgba8") {
     spp = 4; bps = 8;
   } else if (space == "rgb16") {
     spp = 3; bps = 16;
@@ -706,6 +872,17 @@ bool colorspace_by_name (Image& image, const std::string& target_colorspace,
     return false;
   }
 
+  return colorspace_convert(image, spp, bps, threshold);
+}
+
+bool colorspace_convert(Image& image, int spp, int bps, uint8_t threshold)
+{
+  // thru the codec?
+  if (!image.isModified() && image.getCodec())
+    if (spp == 1 && bps >= 8)
+      if (image.getCodec()->toGray(image))
+	return true;
+
   // no image data, e.g. for loading raw images
   if (!image.getRawData()) {
     image.spp = spp;
@@ -722,10 +899,10 @@ bool colorspace_by_name (Image& image, const std::string& target_colorspace,
     colorspace_grayX_to_gray8 (image);
   
   // upscale to 8 bit even for sub byte gray since we have no inter sub conv., yet
-  if (image.bps < 8 && image.bps > bps)
+  if (image.bps < 8 && image.bps != bps)
     colorspace_grayX_to_gray8 (image);
   
-  if (image.bps == 8 && image.spp == 1 && spp == 3)
+  if (image.bps == 8 && image.spp == 1 && spp >= 3)
     colorspace_gray8_to_rgb8 (image);
 
   if (image.bps == 8 && bps == 16)
@@ -735,9 +912,17 @@ bool colorspace_by_name (Image& image, const std::string& target_colorspace,
   if (image.bps == 16 && bps < 16)
     colorspace_16_to_8 (image);
   
-  if (image.spp == 4 && spp < 4 && image.bps == 8) // TODO: might be RGB16
-    colorspace_rgba8_to_rgb8 (image);
- 
+  if (image.spp == 4 && spp < 4 && image.bps == 8) { // TODO: might be RGB16
+    if (spp < 3)
+      colorspace_rgb8_to_gray8 (image, 4);
+    else
+      colorspace_rgba8_to_rgb8 (image);
+  }
+
+  if (image.spp == 3 && spp == 4 && image.bps == 8) { // TODO: might be RGB16
+      colorspace_rgb8_to_rgb8a (image);
+  }
+  
   if (image.spp == 3 && spp == 1) {
     if (image.bps == 8) 
       colorspace_rgb8_to_gray8 (image);
@@ -759,6 +944,10 @@ bool colorspace_by_name (Image& image, const std::string& target_colorspace,
               << spp << ", bps: " << bps
               << " - now at spp: " << image.spp << ", bps: " << image.bps
               << std::endl;
+    image.spp = spp;
+    image.bps = bps;
+    image.resize(image.w, image.h);
+    
     return false;
   }
   return true;
@@ -774,6 +963,7 @@ const char* colorspace_name (Image& image)
     case 8: return "gray8";
     case 16: return "gray16";
     case 24: return "rgb8";
+    case 32: return "rgba8";
     case 48: return "rgb16";
     default: return "";
     }
@@ -835,17 +1025,17 @@ struct brightness_contrast_gamma_template
 	
 	a.getRGB (_r, _g, _b);
 	r = _r, g = _g, b = _b;
-	r /= T::accu::one().v1;
-	g /= T::accu::one().v1;
-	b /= T::accu::one().v1;
+	r /= T::accu::one().v[0];
+	g /= T::accu::one().v[0];
+	b /= T::accu::one().v[0];
 	
 	r = convert (r, brightness, contrast, gamma);
 	g = convert (g, brightness, contrast, gamma);
 	b = convert (b, brightness, contrast, gamma);
 	
-	_r = (typename T::accu::vtype)(r * T::accu::one().v1);
-	_g = (typename T::accu::vtype)(g * T::accu::one().v1);
-	_b = (typename T::accu::vtype)(b * T::accu::one().v1);
+	_r = (typename T::accu::vtype)(r * T::accu::one().v[0]);
+	_g = (typename T::accu::vtype)(g * T::accu::one().v[0]);
+	_b = (typename T::accu::vtype)(b * T::accu::one().v[0]);
 	a.setRGB (_r, _g, _b);
 	it.set(a);
 	++it;
@@ -861,79 +1051,73 @@ void brightness_contrast_gamma (Image& image, double brightness, double contrast
 
 template <typename T>
 struct hue_saturation_lightness_template {
-  void operator() (Image& image, double hue, double saturation, double lightness)
+  void operator() (Image& image, double _hue, double saturation, double lightness)
   {
     T it (image);
-    
-    hue = fmod (hue, 360);
-    if (hue < 0)
-      hue += 360;
-    
-    //saturation = std::max (std::min (saturation, 2.), -2.);
-    //lightness = std::max (std::min (lightness, 2.), -2.);
 
-    typename T::accu a;
-    typename T::accu::vtype _r, _g, _b;
-    double r, g, b, h, s, v;
-    
+    // optimized ONE2 in divisions imprecise shifts if not an FP type
+    const typename T::accu::vtype ONE = T::accu::one().v[0],
+                                  ONE2 = ONE > 1 ? ONE + 1 : ONE;
+
+    // H in degree, S and L [-1, 1], latér scaled to the integer range
+    _hue = fmod (_hue, 360);
+    if (_hue < 0)
+      _hue += 360;
+    const typename T::accu::vtype
+      hue = ONE * _hue / 360;
+
     for (int i = 0; i < image.h * image.w; ++i)
       {
-	// H in degree, S, L [-1, 1]
-	a = *it;
-	
-	a.getRGB (_r, _g, _b);
-	r = _r, g = _g, b = _b;
-	r /= T::accu::one().v1;
-	g /= T::accu::one().v1;
-	b /= T::accu::one().v1;
-	
+	typename T::accu a = *it;	
+	typename T::accu::vtype r, g, b;
+	a.getRGB (r, g, b);
+
 	// RGB to HSV
+	typename T::accu::vtype h, s, v;
 	{
-	  const double min = std::min (std::min (r, g), b);
-	  const double max = std::max (std::max (r, g), b);
-	  const double delta = max - min;
-	  
+	  const typename T::accu::vtype min = std::min (std::min (r, g), b);
+	  const typename T::accu::vtype max = std::max (std::max (r, g), b);
+	  const typename T::accu::vtype delta = max - min;
 	  v = max;
+
 	  if (delta == 0) {
 	    h = 0;
 	    s = 0;
 	  }
 	  else {
-	    s = max == 0 ? 0 : 1. - min / max;
+	    s = max == 0 ? 0 : ONE - (ONE * min / max);
 	    
 	    if (max == r) // yellow - magenta
-	      h = 60. * (g - b) / delta + (g >= b ? 0 : 360);
+	      h = (60*ONE/360) * (g - b) / delta + (g >= b ? 0 : ONE);
 	    else if (max == g) // cyan - yellow
-	      h = 60. * (b - r) / delta + 120;
+	      h = (60*ONE/360) * (b - r) / delta + 120*ONE/360;
 	    else // magenta - cyan
-	      h = 60. * (r - g) / delta + 240;
+	      h = (60*ONE/360) * (r - g) / delta + 240*ONE/360;
 	  }
-	} // end
+	}
 	
+	// hue should only be positive, se we only need to check one overflow
 	h += hue;
-	if (h < 0)
-	  h += 360;
-	else if (h >= 360)
-	  h -= 360;
+	if (h >= ONE)
+	  h -= ONE;
 	
 	// TODO: this might not be accurate, double check, ...
 	s = s + s * saturation;
-	s = std::max (std::min (s, 1.), 0.);
+	s = std::max (std::min (s, ONE), (typename T::accu::vtype)0);
 	
 	v = v + v * lightness;
-	v = std::max (std::min (v, 1.), 0.);
+	v = std::max (std::min (v, ONE), (typename T::accu::vtype)0);
 	
-	
-	// HSV to RGB
+	// back from HSV to RGB
 	{
-	  h /= 60.;
-	  const int i = (int) (floor(h)) % 6;
-	  
-	  const double f = h - i;
-	  const double p = v * (1 - s);
-	  const double q = v * (1 - f * s);
-	  const double t = v * (1 - (1. - f) * s);
+	  const int i = 6 * h / ONE2;
 	  
+	  const typename T::accu::vtype f = 6 * h % ONE2;
+	  // only compute the ones "on-demand" as needed
+	  #define p (v * (ONE - s) / ONE2)
+	  #define q (v * (ONE - f * s / ONE2) / ONE2)
+	  #define t (v * (ONE - (ONE - f) * s / ONE2) / ONE2)
+
 	  switch (i) {
 	  case 0:
 	    r = v;
@@ -966,12 +1150,12 @@ struct hue_saturation_lightness_template {
 	    b = q;
 	    break;
 	  }
-	
+#undef p
+#undef q
+#undef t
 	} // end
-	_r = (typename T::accu::vtype)(r * T::accu::one().v1);
-	_g = (typename T::accu::vtype)(g * T::accu::one().v1);
-	_b = (typename T::accu::vtype)(b * T::accu::one().v1);
-	a.setRGB (_r, _g, _b);
+
+	a.setRGB (r, g, b);
 	it.set(a);
 	++it;
       }
